#include "FreeRTOS.h"

#include "i2c.h"

//
//
//
typedef enum
{
  I2CFLAGS_START         = 0x0001,
  I2CFLAGS_REPEATEDSTART = 0x0002,
  I2CFLAGS_STOP          = 0x0004,
  I2CFLAGS_ADDRESS       = 0x0008,
  I2CFLAGS_WRITEDATA     = 0x0010,
  I2CFLAGS_READDATA      = 0x0020,
}
i2cFlags_e;

typedef enum
{
  I2CMODE_ACK = 0,
  I2CMODE_NACK,
  I2CMODE_READ
}
i2cMode_e;

//
//  Our PCLK is 48Mhz (12Mhz Xtal, PLL x 4, VBPDIV = /1), we want 100Khz SCLK
//
void i2cInit (void)
{
  SCB_PCONP |= SCB_PCONP_PCI2C0;

  PCB_PINSEL0 = (PCB_PINSEL0 & ~(PCB_PINSEL0_P02_MASK | PCB_PINSEL0_P03_MASK)) | (PCB_PINSEL0_P02_SCL0 | PCB_PINSEL0_P03_SDA0);

  I2C0_CONCLR = I2C_CONCLR_MASK;
  I2C0_CONSET = I2C_CONSET_I2EN;
  I2C0_SCLL   = 480;
  I2C0_SCLH   = 480;
}

//
//
//
static i2cErr_e i2cStatus (void)
{
  i2cErr_e status;

  while (!(I2C0_CONSET & I2C_CONSET_SI))
    ;

  if ((status = I2C0_STAT) == I2CERR_BUSERRORx)
    return I2CERR_BUSERROR;

  return status;
}

//
//
//
static i2cErr_e i2cStop (void)
{
  I2C0_CONSET = I2C_CONSET_STO;
  I2C0_CONCLR = I2C_CONCLR_SIC;

  while (I2C0_CONSET & I2C_CONSET_STO)
    ;

  return I2CERR_NONE;
}

//
//
//
static i2cErr_e i2cStart (void)
{
  I2C0_CONSET = I2C_CONSET_STA;

  while (1)
  {
    i2cErr_e status;

    if (((status = i2cStatus ()) == I2CERR_STARTTX) || (status == I2CERR_REPEATEDSTARTTX))
    {
      I2C0_CONCLR = I2C_CONCLR_STAC;
      return I2CERR_NONE;
    }
    else if (status != I2CERR_NOINFO)
    {
      I2C0_CONCLR = I2C_CONCLR_STAC;
      return status;
    }
    else
      I2C0_CONCLR = I2C_CONCLR_SIC;
  }
}

//
//
//
static i2cErr_e i2cRepeatedStart (void)
{
  while (!(I2C0_CONSET & I2C_CONSET_SI))
    ;

  I2C0_CONSET = I2C_CONSET_STA;
  I2C0_CONCLR = I2C_CONCLR_SIC;

  while (1)
  {
    i2cErr_e status;

    if (((status = i2cStatus ()) == I2CERR_STARTTX) || (status == I2CERR_REPEATEDSTARTTX))
    {
      I2C0_CONCLR = I2C_CONCLR_STAC;
      return I2CERR_NONE;
    }
    else if (status != I2CERR_NOINFO)
    {
      I2C0_CONCLR = I2C_CONCLR_STAC;
      return status;
    }
    else
      I2C0_CONCLR = I2C_CONCLR_SIC;
  }
}

//
//
//
static i2cErr_e i2cPutByte (U8 data)
{
  if (!(I2C0_CONSET & I2C_CONSET_SI))
    return I2CERR_BUSY;

  I2C0_DAT = data;
  I2C0_CONCLR = I2C_CONCLR_SIC;

  return I2CERR_NONE;
}

//
//
//
static i2cErr_e i2cGetByte (i2cMode_e mode, U8 *pData)
{
  switch (mode)
  {
    case I2CMODE_ACK :
      I2C0_CONSET = I2C_CONSET_AA;
      I2C0_CONCLR = I2C_CONCLR_SIC;
      break;

    case I2CMODE_NACK :
      I2C0_CONCLR = (I2C_CONCLR_AAC | I2C_CONCLR_SIC);
      break;

    case I2CMODE_READ :
      {
        if (!(I2C0_CONSET & I2C_CONSET_SI))
          return I2CERR_EMPTY;

        *pData = (U8) I2C0_DAT;
      }
      break;
  }

  return I2CERR_NONE;
} 

//
//
//
static i2cErr_e i2cWriteBufferEx (U8 address, U8 *buffer, U32 bufferLength, i2cFlags_e flags)
{
  U32 i;
  i2cErr_e status;

  if (flags & I2CFLAGS_START)
  {
    if ((status = i2cStart ()) != I2CERR_NONE)
    {
      i2cStop ();
      return status;
    }
  }
  else if (flags & I2CFLAGS_REPEATEDSTART)
  {
    if ((status = i2cRepeatedStart ()) != I2CERR_NONE)
    {
      i2cStop ();
      return status;
    }
  }

  if (flags & I2CFLAGS_ADDRESS)
  {
    do
      if (((status = i2cPutByte (address & ~0x01)) != I2CERR_NONE) && (status != I2CERR_BUSY))
        return status;
    while (status == I2CERR_BUSY);
  }

  if (flags & I2CFLAGS_WRITEDATA)
  {
    for (i = 0; i < bufferLength; i++, buffer++)
    {
      while (1)
      {
        if (((status = i2cStatus ()) == I2CERR_SLAWTX_ACKRX) || (status == I2CERR_DATTX_ACKRX))
        {
          do
            if (((status = i2cPutByte (*buffer)) != I2CERR_NONE) && (status != I2CERR_BUSY))
              return status;
          while (status == I2CERR_BUSY);

          break;
        }
        else if (status != I2CERR_NOINFO)
        {
          i2cStop ();
          return status;
        }
      }
    }
  }

  if (flags & I2CFLAGS_STOP)
  {
    while (1)
    {
      if (((status = i2cStatus ()) == I2CERR_SLAWTX_ACKRX) || (status == I2CERR_DATTX_ACKRX))
      {
        i2cStop ();
        return I2CERR_NONE;
      }
      else if (status != I2CERR_NOINFO)
      {
        i2cStop ();
        return status;
      }
    }
  }

  return I2CERR_NONE;
}

//
//
//
static i2cErr_e i2cReadBufferEx (U8 address, U8 *buffer, U32 bufferLength, i2cFlags_e flags)
{
  U32 i;
  i2cErr_e status;

  if (flags & I2CFLAGS_START)
  {
    if ((status = i2cStart ()) != I2CERR_NONE)
    {
      i2cStop ();
      return status;
    }
  }
  else if (flags & I2CFLAGS_REPEATEDSTART)
  {
    if ((status = i2cRepeatedStart ()) != I2CERR_NONE)
    {
      i2cStop ();
      return status;
    }
  }

  if (flags & I2CFLAGS_ADDRESS)
  {
    do
      if (((status = i2cPutByte (address | 0x01)) != I2CERR_NONE) && (status != I2CERR_BUSY))
        return status;
    while (status == I2CERR_BUSY);
  }

  if (flags & I2CFLAGS_READDATA)
  {
    for (i = 0; i < bufferLength; i++, buffer++)
    {
      while (1)
      {
        if (((status = i2cStatus ()) == I2CERR_SLARTX_ACKRX) || (status == I2CERR_SLARTX_NACKRX) || (status == I2CERR_DATRX_ACKTX))
        {
          i2cGetByte ((i != bufferLength - 1) ? I2CMODE_ACK : I2CMODE_NACK, NULL);

          do
            status = i2cGetByte (I2CMODE_READ, buffer);
          while (status == I2CERR_EMPTY);

          break;
        }
        else if (status != I2CERR_NOINFO)
        {
          i2cStop ();
          return status;
        }
      }
    }
  }

  if (flags & I2CFLAGS_STOP)
    i2cStop ();

  return I2CERR_NONE;
}

//
//
//
i2cErr_e i2cWriteBuffer (U8 address, U8 *buffer, U32 bufferLength)
{
  return i2cWriteBufferEx (address, buffer, bufferLength, I2CFLAGS_START | I2CFLAGS_ADDRESS | I2CFLAGS_WRITEDATA | I2CFLAGS_STOP);
}

i2cErr_e i2cReadBuffer (U8 address, U8 *buffer, U32 bufferLength)
{
  return i2cReadBufferEx (address, buffer, bufferLength, I2CFLAGS_START | I2CFLAGS_ADDRESS | I2CFLAGS_READDATA | I2CFLAGS_STOP);
}

i2cErr_e i2cWriteReadBuffer (U8 address, U8 *buffer, U32 putLength, U32 getLength)
{
  i2cErr_e status;

  if ((status = i2cWriteBufferEx (address, buffer, putLength, I2CFLAGS_START | I2CFLAGS_ADDRESS | I2CFLAGS_WRITEDATA)) != I2CERR_NONE)
    return status;

  return i2cReadBufferEx (address, buffer, getLength, I2CFLAGS_REPEATEDSTART | I2CFLAGS_ADDRESS | I2CFLAGS_READDATA | I2CFLAGS_STOP);
}
